/************************************************************************
Demo Code for Designing_Visual_Editors_in_Qt Article

Copyright (C) 2004-2007  Prashanth N Udupa, (prashanth @ vcreatelogic . com)
        VCreate Logic Private Limited, Bangalore

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
version 2 as published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License Version 2 for more details.

You should have received a copy of the GNU General Public License
Version 2 along with this program; if not, write to the Free
Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA
************************************************************************/

#include "CWrapperCanvasView.h"
#include "CWrapperClassLayer.h"

#include <QCursor>
#include <QMatrix>
#include <QPainter>
#include <QMouseEvent>
#include <QWheelEvent>
#include <QApplication>

struct CWrapperCanvasViewData
{
    CWrapperCanvasViewData() : canvas(0), zoomLevel(0), zoomValue(1.0f) { }

    CWrapperCanvas* canvas;
    CWrapperCanvasView* canvasView;
    int zoomLevel;
    qreal zoomValue;

    qreal calculateScaleFor(int zoomLevel) const;
    int calculateZoomLevelFor(qreal scale) const;
};

CWrapperCanvasView::CWrapperCanvasView(QWidget* parent)
:QGraphicsView(parent)
{
    d = new CWrapperCanvasViewData;
    d->canvasView = this;

    setWrapperCanvas(&CWrapperCanvas::instance());
    setRenderHint(QPainter::Antialiasing, true);
}

CWrapperCanvasView::~CWrapperCanvasView()
{
    delete d;
}

void CWrapperCanvasView::setWrapperCanvas(CWrapperCanvas* canvas)
{
    if(d->canvas == canvas)
        return;

    d->canvas = canvas;
    QGraphicsView::setScene(canvas);
}

CWrapperCanvas* CWrapperCanvasView::wrapperCanvas() const
{
    return d->canvas;
}

int CWrapperCanvasView::zoomLevel() const
{
    return d->zoomLevel;
}

double CWrapperCanvasView::zoomScale() const
{
    return d->zoomValue;
}

void CWrapperCanvasView::setZoomLevel(int zoomLevel)
{
    if(d->zoomLevel == zoomLevel)
        return;

    // Clip zoom level between -10 and +10.
    if(zoomLevel < -10)
        zoomLevel = -10;
    else if(zoomLevel > 10)
        zoomLevel = 10;

    if(d->zoomLevel == zoomLevel)
        return;

    d->zoomLevel = zoomLevel;
    d->zoomValue = d->calculateScaleFor(zoomLevel);

    // Set the zoom to the graphics view.
    QApplication::setOverrideCursor(Qt::WaitCursor);
    QMatrix matrix;
    matrix.scale(d->zoomValue, d->zoomValue);
    setMatrix(matrix);
    QApplication::restoreOverrideCursor();

    emit zoomLevelChanged(d->zoomLevel);
    emit zoomScaleChanged(d->zoomValue);
}

void CWrapperCanvasView::zoomIn()
{
    setZoomLevel(d->zoomLevel+1);
}

void CWrapperCanvasView::zoomOut()
{
    setZoomLevel(d->zoomLevel-1);
}

void CWrapperCanvasView::zoomFull()
{
    setZoomLevel(10);
}

void CWrapperCanvasView::zoomOne()
{
    setZoomLevel(0);
}

void CWrapperCanvasView::zoomFit()
{
    QRectF iRect = d->canvas->itemsBoundingRect();
    fitInView(iRect, Qt::KeepAspectRatio);

    qreal scale = matrix().m11();
    d->zoomLevel = d->calculateZoomLevelFor(scale);

    emit zoomLevelChanged(d->zoomLevel);
    emit zoomScaleChanged(scale);
}

void CWrapperCanvasView::printToPDF(const QString fileName)
{
    // TBD
    Q_UNUSED(fileName)
}

void CWrapperCanvasView::mousePressEvent(QMouseEvent* e)
{
    QGraphicsView::mousePressEvent(e);
}

void CWrapperCanvasView::mouseMoveEvent(QMouseEvent* e)
{
    QGraphicsView::mouseMoveEvent(e);
}

void CWrapperCanvasView::mouseReleaseEvent(QMouseEvent* e)
{
    QGraphicsView::mouseReleaseEvent(e);
}

void CWrapperCanvasView::resizeEvent(QResizeEvent* re)
{
    QGraphicsView::resizeEvent(re);

    if(d->zoomLevel != 0)
    {
        int zl = d->zoomLevel;
        d->zoomLevel = 0;
        setZoomLevel(zl);
    }
}

void CWrapperCanvasView::wheelEvent(QWheelEvent * e)
{
    if(e->delta() > 0)
        zoomIn();
    else
        zoomOut();
    e->accept();
}

void CWrapperCanvasView::drawBackground(QPainter* paint, const QRectF & rect)
{
    if(!isEnabled())
        return;

    QMatrix matrix = paint->matrix();
    double dist = 10.0f; // 10 pixels distance between grid points.

    int gridXCount = int(sceneRect().width() / dist) + 1;
    int gridYCount = int(sceneRect().height() / dist) + 1;

    // Figure out the part of the grid that actually needs to
    // be redrawn.
    int gridXStart = int((rect.left() - sceneRect().left()) / dist);
    int gridYStart = int((rect.top() - sceneRect().top()) / dist);

    if(gridXStart < 0)
        gridXStart = 0;
    if(gridYStart < 0)
        gridYStart = 0;

    paint->fillRect(sceneRect(), Qt::white);

    paint->save();
    paint->resetMatrix();

    double x = sceneRect().left();
    double y = sceneRect().top();

    for(int i=gridXStart; i<gridXCount; i++)
    {
        for(int j=gridYStart; j<gridYCount; j++)
        {
            x = sceneRect().left() + double(i) * dist;
            y = sceneRect().top() + double(j) * dist;
            paint->drawPoint(matrix.map(QPointF(x, y)));
        }
    }

    paint->restore();

    paint->drawRect(sceneRect());
}

qreal CWrapperCanvasViewData::calculateScaleFor(int zoomLevel) const
{
    if(!canvas)
        return 1.0f;

    qreal canvasWidth = canvas->width();
    qreal canvasHeight = canvas->height();

    qreal canvasSize = canvasWidth > canvasHeight ? canvasWidth : canvasHeight;
    qreal viewSize = canvasWidth > canvasHeight ?
                        qreal(canvasView->viewport()->width()) :
                        qreal(canvasView->viewport()->height());

    // At zoom level -10 canvasSize must become viewSize;
    qreal zoomOutFactor = (viewSize*0.95f / canvasSize);

    // At zoom level +10 canvasSize must be 3 times its size.
    qreal zoomInFactor = 3.0f / 10.0f;

    // Calculate the zoom
    qreal zoom = 1.0f;
    if(zoomLevel < 0)
        zoom = 1.0f + qreal(zoomLevel) * (1.0f - zoomOutFactor) / 10.0f;
    else
        zoom = 1.0f + zoomInFactor * qreal(zoomLevel);

    return zoom;
}

int CWrapperCanvasViewData::calculateZoomLevelFor(qreal scale) const
{
    if(!canvas)
        return 0;

    qreal canvasWidth = canvas->width();
    qreal canvasHeight = canvas->height();

    qreal canvasSize = canvasWidth > canvasHeight ? canvasWidth : canvasHeight;
    qreal viewSize = canvasWidth > canvasHeight ?
                        qreal(canvasView->viewport()->width()) :
                        qreal(canvasView->viewport()->height());

    // At zoom level -10 canvasSize must become viewSize;
    qreal zoomOutFactor = (viewSize*0.95f / canvasSize);

    // At zoom level +10 canvasSize must be 3 times its size.
    qreal zoomInFactor = 3.0f / 10.0f;

    qreal zoomLevel = 1.0f;

    if(scale < 1)
        zoomLevel = (scale - 1.0f) * 10.0f / (1.0f - zoomOutFactor);
    else
        zoomLevel = (scale - 1.0f) / zoomInFactor;

    return int(zoomLevel);
}

