/************************************************************************
Demo Code for Designing_Visual_Editors_in_Qt Article

Copyright (C) 2004-2007  Prashanth N Udupa, (prashanth @ vcreatelogic . com)
        VCreate Logic Private Limited, Bangalore

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
version 2 as published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License Version 2 for more details.

You should have received a copy of the GNU General Public License
Version 2 along with this program; if not, write to the Free
Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA  02111-1307  USA
************************************************************************/

#include "CWrapperClassLayer.h"
#include "CWrapperClassSelectionList.h"
#include "CWrapperFactory.h"
#include "vtkObject.h"

#include <QMessageBox>
#include <QLineF>
#include <QMenu>
#include <QEvent>
#include <QPainter>
#include <QApplication>
#include <QLinearGradient>
#include <QStyleOptionGraphicsItem>
#include <QGraphicsSceneMouseEvent>
#include <QGraphicsSceneContextMenuEvent>

////////////////////////////////////////////////////////////////////////////////////////////////////
/// CWrapperLinkDesc
////////////////////////////////////////////////////////////////////////////////////////////////////
CWrapperLinkDesc::CWrapperLinkDesc()
{
    linkName = "NoName";
    linkDataType = -1; // code for invalid type
    acceptMultipleInputs = false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// CWrapperLinkData
////////////////////////////////////////////////////////////////////////////////////////////////////
CWrapperLinkData::CWrapperLinkData()
{
    data.dataObject = 0; // all pointers are zero now.
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// CWrapper
////////////////////////////////////////////////////////////////////////////////////////////////////
typedef QList<CWrapperLink*> CWrapperLinkList;

struct CWrapperData
{
    CWrapperData() : vtkObjectPtr(0) { }
    ~CWrapperData() {
        if(vtkObjectPtr)
            vtkObjectPtr->Delete();
    }

    vtkObject* vtkObjectPtr;
    QMap<int, CWrapperLinkList> inputLinks;
    QMap<int, CWrapperLinkList> outputLinks;
};

bool CWrapper::link(CWrapper* output, int outLine, CWrapper* input, int inLine, QString* errMsg)
{
    // Check all input parameters
    if(!output || !input || outLine<0 || inLine<0)
    {
        if(errMsg)
            *errMsg = "Invalid parameters passed to link";
        return false;
    }

    // Check the state of the wrappers taking part in the link
    if(!output->getVtkObject() || !input->getVtkObject())
    {
        if(errMsg)
            *errMsg = "Invalid Wrapper State";
        return false;
    }

    // find out if a similar link already exists
    CWrapperLink* wrapperLink = CWrapperCanvas::instance().link(output, outLine, input, inLine);
    if(wrapperLink)
    {
        if(errMsg)
            *errMsg = "Link already exists";
        return false;
    }

    // Check for link type compatibility
    CWrapperLinkDesc inputDesc = input->inputLinkDesc(inLine);
    CWrapperLinkDesc outputDesc = output->outputLinkDesc(outLine);
    if(inputDesc.linkDataType != outputDesc.linkDataType)
    {
        if(errMsg)
            *errMsg = "Incompatible link types. Cannot make a connection";
        return false;
    }

    if(!inputDesc.acceptMultipleInputs && input->hasInput(inLine))
    {
        if(errMsg)
            *errMsg = "Cannot accept more than one input";
        return false;
    }

    // Perform the link
    CWrapperLinkData linkData;
    linkData.linkDesc = output->outputLinkDesc(outLine);
    bool success = output->getOutput(outLine, linkData);
    if(!success || !linkData.data.dataObject)
    {
        if(errMsg)
            *errMsg = "Could not fetch output";
        return false;
    }
    success = input->setInput(inLine, linkData);
    if(!success)
    {
        if(errMsg)
            *errMsg = "Cannot set input";
        return false;
    }

    // Now that the link is successfully established create a CWrapperLink
    // for the newly created link
    wrapperLink = new CWrapperLink(output, outLine, input, inLine);

    // add the link to the input and output wrappers.
    input->addLink(wrapperLink);
    output->addLink(wrapperLink);

    // All done.
    return true;
}

bool CWrapper::unlink(CWrapper* output, int outLine, CWrapper* input, int inLine, QString* errMsg)
{
    // Check all input parameters
    if(!output || !input || outLine<0 || inLine<0)
    {
        if(errMsg)
            *errMsg = "Invalid parameters passed to link";
        return false;
    }

    // Check the state of the wrappers taking part in the link
    if(!output->getVtkObject() || !input->getVtkObject())
    {
        if(errMsg)
            *errMsg = "Invalid Wrapper State";
        return false;
    }

    // Find out the link
    CWrapperLink* wrapperLink = CWrapperCanvas::instance().link(output, outLine, input, inLine);
    if(!wrapperLink)
    {
        if(errMsg)
            *errMsg = "Invalid link";
        return false;
    }

    // remove the link now.
    CWrapperLinkData linkData;
    bool success = output->getOutput(outLine, linkData);
    if(!success || !linkData.data.dataObject)
    {
        if(errMsg)
            *errMsg = "Could not fetch output";
        return false;
    }

    success = input->removeInput(inLine, linkData);
    if(!success)
    {
        if(errMsg)
            *errMsg = "Cannot remove input";
        return false;
    }

    // remove the link from both input and output wrapper.
    input->removeLink(wrapperLink);
    output->removeLink(wrapperLink);

    // Now delete the wrapper link itself.
    if(!wrapperLink->inDestructor())
    {
        wrapperLink->setAutoUnlink(false);
        delete wrapperLink;
    }

    // All done.
    return true;
}

CWrapper::CWrapper()
{
    d = new CWrapperData;

    static int count = 0;
    setObjectName( QString("object%1").arg(count++) );
    setRect(-100, -30, 200, 60);

    CWrapperCanvas::instance().addItem(this);
    CWrapperCanvas::instance().addWrapper(this);
    setFlags(ItemIsMovable|ItemIsSelectable|ItemIsFocusable);

    setZValue(1.0f);
}

CWrapper::~CWrapper()
{
    CWrapperCanvas::instance().removeWrapper(this);

    // Use deleteLater().
    delete d;
}

vtkObject* CWrapper::getVtkObject() const
{
    return d->vtkObjectPtr;
}

int CWrapper::inputLinkCount() const
{
    return 0;
}

CWrapperLinkDesc CWrapper::inputLinkDesc(int index) const
{
    Q_UNUSED(index);
    return CWrapperLinkDesc();
}

QRectF CWrapper::inputLinkRect(int index) const
{
    // each rectangle is 7 pixels wide. We will need
    // 5 pixels gap between consequetive rectangles.
    qreal reqSpace = 7.0f * qreal(2*inputLinkCount()-1);

    // figure out the x of the first link.
    qreal x = rect().left() + (rect().width() - reqSpace) / 2.0f;

    // figure out the x of the index's link.
    x += 12.0f * qreal(index);

    // Construct a rectangle around this.
    return QRectF(x, rect().top(), 7.0f, 7.0f);
}

QList<CWrapperLink*> CWrapper::inputLinksAt(int index) const
{
    if(d->inputLinks.contains(index))
        return d->inputLinks[index];
    return QList<CWrapperLink*>();
}

int CWrapper::outputLinkCount() const
{
    return 0;
}

CWrapperLinkDesc CWrapper::outputLinkDesc(int index) const
{
    Q_UNUSED(index)
    return CWrapperLinkDesc();
}

QRectF CWrapper::outputLinkRect(int index) const
{
    // each rectangle is 7 pixels wide. We will need
    // 5 pixels gap between consequetive rectangles.
    qreal reqSpace = 7.0f * qreal(2*outputLinkCount()-1);

    // figure out the x of the first link.
    qreal x = rect().left() + (rect().width() - reqSpace) / 2.0f;

    // figure out the x of the index's link.
    x += 12.0f * qreal(index);

    // Construct a rectangle around this.
    return QRectF(x, rect().bottom()-7.0f, 7.0f, 7.0f);
}

QList<CWrapperLink*> CWrapper::outputLinksAt(int index) const
{
    if(d->outputLinks.contains(index))
        return d->outputLinks[index];
    return QList<CWrapperLink*>();
}

void CWrapper::removeInputLinks()
{
    QMap<int, CWrapperLinkList>::iterator it = d->inputLinks.begin();
    QMap<int, CWrapperLinkList>::iterator end = d->inputLinks.end();
    while(it != end)
    {
        qDeleteAll(it.value());
        ++it;
    }
    d->inputLinks.count();
}

void CWrapper::removeOutputLinks()
{
    QMap<int, CWrapperLinkList>::iterator it = d->outputLinks.begin();
    QMap<int, CWrapperLinkList>::iterator end = d->outputLinks.end();
    while(it != end)
    {
        qDeleteAll(it.value());
        ++it;
    }
    d->outputLinks.count();
}

void CWrapper::removeAllLinks()
{
    removeInputLinks();
    removeOutputLinks();
}

bool CWrapper::hasInput(int index)
{
    Q_UNUSED(index);
    return false;
}

bool CWrapper::setInput(int index, const CWrapperLinkData & linkData)
{
    Q_UNUSED(index);
    Q_UNUSED(linkData);
    return false;
}

bool CWrapper::removeInput(int index, const CWrapperLinkData & linkData)
{
    Q_UNUSED(index);
    Q_UNUSED(linkData);
    return false;
}

bool CWrapper::getOutput(int index, CWrapperLinkData & output)
{
    Q_UNUSED(index);
    Q_UNUSED(output);
    return false;
}

void CWrapper::setVtkObject(vtkObject* object)
{
    d->vtkObjectPtr = object;
}

bool CWrapper::event(QEvent* event)
{
    switch(event->type())
    {
    case QEvent::DeferredDelete:
        removeAllLinks();
        break;
    default:
        break;
    }

    return QObject::event(event);
}

void CWrapper::paint(QPainter* p, const QStyleOptionGraphicsItem* opt, QWidget* w)
{
    Q_UNUSED(w)

    QLinearGradient grad(rect().topLeft(), rect().bottomLeft());
    if(isSelected())
    {
        grad.setColorAt(0.0f, opt->palette.mid().color());
        grad.setColorAt(0.5f, opt->palette.button().color());
        grad.setColorAt(1.0f, opt->palette.mid().color());
    }
    else
    {
        grad.setColorAt(0.0f, opt->palette.light().color());
        grad.setColorAt(0.5f, opt->palette.button().color());
        grad.setColorAt(1.0f, opt->palette.light().color());
    }

    p->fillRect(rect(), grad);
    p->drawRect(rect());

    for(int i=0; i<inputLinkCount(); i++)
    {
        QRectF ir = inputLinkRect(i);
        p->fillRect(ir, opt->palette.dark());
        p->drawRect(ir);
    }

    for(int i=0; i<outputLinkCount(); i++)
    {
        QRectF orect = outputLinkRect(i);
        p->fillRect(orect, opt->palette.dark());
        p->drawRect(orect);
    }

    QRectF nameRect(rect().left(), rect().top(), rect().width(), rect().height()/2.0f);
    QRectF classRect(nameRect.left(), nameRect.bottom(), nameRect.width(), nameRect.height());
    p->drawText(nameRect, Qt::AlignCenter, name());
    if(d->vtkObjectPtr)
        p->drawText(classRect, Qt::AlignCenter, QString("%1").arg(d->vtkObjectPtr->GetClassName()));
    else
        p->drawText(classRect, Qt::AlignCenter, "Unknown Class");
}

QVariant CWrapper::itemChange(GraphicsItemChange change, const QVariant & value)
{
    if(change == ItemPositionChange)
    {
        QList<CWrapperLinkList> list = d->inputLinks.values();
        list += d->outputLinks.values();
        for(int i=0; i<list.count(); i++)
        {
            CWrapperLinkList linkList = list[i];
            for(int j=0; j<linkList.count(); j++)
                linkList[j]->updateLink();
        }
    }
    else if(change == ItemSelectedChange)
    {
        if(value.toBool())
            emit CWrapperCanvas::instance().wrapperSelected(this);
        else
            emit CWrapperCanvas::instance().wrapperSelected(0);
    }
    return value;
}

void CWrapper::addLink(CWrapperLink* link)
{
    if(!link)
        return;

    if(link->input() == this)
        d->inputLinks[link->inputLinkIndex()].push_back(link);
    else if(link->output() == this)
        d->outputLinks[link->outputLinkIndex()].push_back(link);
}

void CWrapper::removeLink(CWrapperLink* link)
{
    if(!link)
        return;

    if(link->input() == this)
        d->inputLinks[link->inputLinkIndex()].removeAll(link);
    else if(link->output() == this)
        d->outputLinks[link->outputLinkIndex()].removeAll(link);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// CWrapperLink
////////////////////////////////////////////////////////////////////////////////////////////////////
struct CWrapperLinkData2
{
    CWrapperLinkData2() : autoUnlink(true), inDestructor(false),
        input(0), output(0), inLine(0), outLine(0) { }

    bool autoUnlink;
    bool inDestructor;
    CWrapper* input;
    CWrapper* output;
    int inLine;
    int outLine;
};

CWrapperLink::CWrapperLink(CWrapper* output, int outLine, CWrapper* input, int inLine)
{
    d = new CWrapperLinkData2;
    d->output = output;
    d->input = input;
    d->outLine = outLine;
    d->inLine = inLine;
    updateLink();

    CWrapperCanvas::instance().addItem(this);
    CWrapperCanvas::instance().addLink(this);
    setZValue(2.0f);
    setPen(QPen(Qt::black, 2.0f));
}

CWrapperLink::~CWrapperLink()
{
    d->inDestructor = true;
    if(d->autoUnlink)
        CWrapper::unlink(d->output, d->outLine, d->input, d->inLine);
    CWrapperCanvas::instance().removeLink(this);
    delete d;
}

bool CWrapperLink::inDestructor() const
{
    return d->inDestructor;
}

CWrapper* CWrapperLink::output() const
{
    return d->output;
}

int CWrapperLink::outputLinkIndex() const
{
    return d->outLine;
}

CWrapper* CWrapperLink::input() const
{
    return d->input;
}

int CWrapperLink::inputLinkIndex() const
{
    return d->inLine;
}

void CWrapperLink::updateLink()
{
    if(!d->output || !d->input || d->inLine < 0 || d->outLine < 0)
        setLine(QLineF());
    else
    {
        QRectF ir = d->input->inputLinkRect(d->inLine);
        QRectF orect = d->output->outputLinkRect(d->outLine);
        ir = d->input->mapToScene(ir).boundingRect();
        orect = d->output->mapToScene(orect).boundingRect();
        setLine(orect.center().x(), orect.center().y(), ir.center().x(), ir.center().y());
    }
}

void CWrapperLink::setAutoUnlink(bool val)
{
    d->autoUnlink = val;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// CWrapperCanvas
////////////////////////////////////////////////////////////////////////////////////////////////////
struct CWrapperCanvasData
{
    CWrapperCanvasData() : ctorLine(0),
        output(0), outLine(-1), input(0), inLine(-1) { }

    QList<CWrapper*> wrapperList;
    QList<CWrapperLink*> wrapperLinkList;

    QGraphicsLineItem* ctorLine;
    CWrapper* output;
    int outLine;
    CWrapper* input;
    int inLine;
    bool inDestructor;
};

CWrapperCanvas & CWrapperCanvas::instance()
{
    static CWrapperCanvas* theInstance = new CWrapperCanvas(qApp);
    return *theInstance;
}

CWrapperCanvas::CWrapperCanvas(QObject* parent)
:QGraphicsScene(parent)
{
    d = new CWrapperCanvasData;
    setSceneRect(-1000, -1000, 1000, 1000);

    QObject::connect(qApp, SIGNAL(aboutToQuit()), this, SLOT(appAboutToQuit()));
}

CWrapperCanvas::~CWrapperCanvas()
{
    d->inDestructor = true;
    appAboutToQuit();
    delete d;
    d = 0;
}

int CWrapperCanvas::wrapperCount() const
{
    return d->wrapperList.count();
}

CWrapper* CWrapperCanvas::wrapper(int index) const
{
    if(index < 0 || index >= d->wrapperList.count())
        return d->wrapperList[index];
    return 0;
}

CWrapper* CWrapperCanvas::wrapper(const QString name) const
{
    for(int i=0; i<d->wrapperList.count(); i++)
        if(d->wrapperList[i]->name() == name)
            return d->wrapperList[i];
    return 0;
}

int CWrapperCanvas::linkCount() const
{
    return d->wrapperLinkList.count();
}

CWrapperLink* CWrapperCanvas::link(int index) const
{
    if(index < 0 || index >= d->wrapperLinkList.count())
        return d->wrapperLinkList[index];
    return 0;
}

CWrapperLink* CWrapperCanvas::link(CWrapper* output, int outLine, CWrapper* input, int inLine)
{
    for(int i=0; i<d->wrapperLinkList.count(); i++)
    {
        CWrapperLink* link = d->wrapperLinkList[i];
        if(link->input() == input && link->output() == output &&
           link->inputLinkIndex() == inLine && link->outputLinkIndex() == outLine)
            return link;
    }
    return 0;
}

QList<CWrapperLink*> CWrapperCanvas::link(CWrapper* output, CWrapper* input)
{
    QList<CWrapperLink*> ret;

    for(int i=0; i<d->wrapperLinkList.count(); i++)
        {
            CWrapperLink* link = d->wrapperLinkList[i];
            if(link->input() == input && link->output() == output)
                ret.push_back(link);
    }

    return ret;
}

CWrapper* CWrapperCanvas::wrapperAt(const QPointF pos, int* inputLink, int* outputLink) const
{
    // pos is in scene coordinates.
    QGraphicsItem* item = itemAt(pos);
    if(!item)
        return 0;

    CWrapper* wrapper = qgraphicsitem_cast<CWrapper*>(item);
    if(!wrapper)
        return 0;

    if(!inputLink && !outputLink)
        return wrapper;

    QPointF itemPos = item->mapFromScene(pos);

    if(inputLink && itemPos.y() < wrapper->rect().center().y())
    {
        *inputLink = -1;
        for(int i=0; i<wrapper->inputLinkCount(); i++)
        {
            QRectF ir = wrapper->inputLinkRect(i);
            if(ir.contains(itemPos))
            {
                *inputLink = i;
                break;
            }
        }
    }

    if(outputLink && itemPos.y() > wrapper->rect().center().y())
    {
        *outputLink = -1;
        for(int i=0; i<wrapper->outputLinkCount(); i++)
        {
            QRectF orect = wrapper->outputLinkRect(i);
            if(orect.contains(itemPos))
            {
                *outputLink = i;
                break;
            }
        }
    }

    return wrapper;
}

void CWrapperCanvas::newCanvas()
{
    QList<CWrapperLink*> links = d->wrapperLinkList;
    QList<CWrapper*> wrappers = d->wrapperList;

    for(int i=0; i<links.count(); i++)
        delete links[i];
    for(int i=0; i<wrappers.count(); i++)
        delete wrappers[i];
}

void CWrapperCanvas::mousePressEvent(QGraphicsSceneMouseEvent* e)
{
    d->output = wrapperAt(e->scenePos(), 0, &d->outLine);
    if(!d->output || d->outLine < 0)
    {
        QGraphicsScene::mousePressEvent(e);
        return;
    }

    d->ctorLine = new QGraphicsLineItem(0, this);
    d->ctorLine->setZValue(3.0f);
    d->ctorLine->setPen(QPen(Qt::blue, 2.0f));
    QRectF orect = d->output->outputLinkRect(d->outLine);
    orect = d->output->mapToScene(orect).boundingRect();
    d->ctorLine->setLine(orect.center().x(), orect.center().y(), e->scenePos().x(), e->scenePos().y());
    d->ctorLine->show();
}

void CWrapperCanvas::mouseMoveEvent(QGraphicsSceneMouseEvent* e)
{
    if(d->ctorLine)
    {
        QLineF line = d->ctorLine->line();
        d->ctorLine->setLine(line.x1(), line.y1(), e->scenePos().x(), e->scenePos().y());
    }
    else
        QGraphicsScene::mouseMoveEvent(e);
}

void CWrapperCanvas::mouseReleaseEvent(QGraphicsSceneMouseEvent* e)
{
    if(!d->ctorLine)
    {
        QGraphicsScene::mouseReleaseEvent(e);
        return;
    }

    delete d->ctorLine;
    d->ctorLine = 0;

    d->input = wrapperAt(e->scenePos(), &d->inLine, 0);
    if(!d->input || d->inLine < 0)
        return;

    QString errMsg;
    bool val = CWrapper::link(d->output, d->outLine, d->input, d->inLine, &errMsg);
    if(!val)
        QMessageBox::information(e->widget(), "Link Error", errMsg);

    d->output = 0; 
    d->outLine = -1;
    d->input = 0;
    d->inLine = -1;
}

void CWrapperCanvas::contextMenuEvent(QGraphicsSceneContextMenuEvent* e)
{
    CWrapper* w = wrapperAt(e->scenePos());
    if(!w)
    {
        QGraphicsScene::contextMenuEvent(e);
        return;
    }

    QMenu menu;
    QAction* action = menu.addAction("Delete");
    QAction* result = menu.exec(QCursor::pos());
    if(result == action)
    {
        w->hide();
        w->deleteLater();
    }
}

void CWrapperCanvas::dragEnterEvent(QGraphicsSceneDragDropEvent* de)
{
    dragMoveEvent(de);
}

void CWrapperCanvas::dragMoveEvent(QGraphicsSceneDragDropEvent* de)
{
    const QMimeData* mimeData = de->mimeData();
    if(!CWrapperClassSelectionListDragObject::canDecode(mimeData))
        return;
    de->setAccepted(true);
    de->acceptProposedAction();
}

void CWrapperCanvas::dropEvent(QGraphicsSceneDragDropEvent* de)
{
    const QMimeData* mimeData = de->mimeData();
    if(!CWrapperClassSelectionListDragObject::canDecode(mimeData))
        return;

    QPointF pos = de->scenePos();
    if(!sceneRect().contains(pos))
        return;

    // Good, we can accept the object being dragged.
    QString className;
    CWrapperClassSelectionListDragObject::decode(mimeData, className);
    if(className.isEmpty())
        return;

    CWrapper* wrapper = CWrapperFactory::instance().createWrapper(className);
    if(wrapper)
        wrapper->setPos(pos);

    // All done !!
    de->acceptProposedAction();
}


void CWrapperCanvas::addWrapper(CWrapper* wrapper)
{
    if(!wrapper)
        return;
    d->wrapperList.push_back(wrapper);
    wrapper->show();
}

void CWrapperCanvas::removeWrapper(CWrapper* wrapper)
{
    if(!wrapper)
        return;
    d->wrapperList.removeAll(wrapper);
}

void CWrapperCanvas::addLink(CWrapperLink* link)
{
    if(!link)
        return;
    d->wrapperLinkList.push_back(link);
    link->show();
}

void CWrapperCanvas::removeLink(CWrapperLink* link)
{
    if(!link)
        return;
    d->wrapperLinkList.removeAll(link);
}

void CWrapperCanvas::appAboutToQuit()
{
    newCanvas();
}


